/*
 *	Programme de Mandelbrot sequentiel  :
 *	cree un fichier pgm à visualiser avec eog
 *
 *	Compilation : g++ -g -Wall -std=c++11 image.c mandel-list.cpp -o mandel.exx
 *
 *  Quelques jolies fractales
 *      ./mandel.ex 2000 2000 0.35 0.355 0.353 0.358 200
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 500
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 300
 *      ./mandel.ex 2000 2000 -1.48478 0.00006 -1.48440 0.00044 100
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>	/* chronometrage */
#include "image.h"

// *********************************************************************
//
//  Ne pas modifier ici
//  Le début du code sert à gérer la liste des tâches d'une façon qui
//  n'est pas adaptée au threads. Vous n'avez pas à la comprendre car
//  votre but est de "protéger" ce code dangereux Vous ne devez modifier
//  que la boucle qui lance les calculs dans la fonction mail.
//
// *********************************************************************
// description du travail
typedef struct {
  // les coordonnée de la zone à afficher
  double xmin;
  double xmax;
  double ymin;
  double ymax;
  // la dimension de l'image
  int dimX;
  int dimY;
  // la profondeur de calcul
  int prof;
} descr;

typedef struct {
  int *tab;
  int nbtaches;
  int debut;
} lt;

lt* new_liste_taches(int nbtaches) {
  int i;
  lt *res = (lt *) malloc(sizeof(lt));

  res->nbtaches = nbtaches;
  res->debut = 0;
  res->tab = (int *) malloc(nbtaches*sizeof(int));
  for (i=0; i<nbtaches; i++) {
    res->tab[i] = i;
  }
  return res;
}

void
delete_liste_taches(lt *l) {
  free(l->tab);
  free(l);
}

// fonction ui retourne une tache (ou -1 si la liste est vide)
int get_tache(lt *l) {
  if (l->debut == l->nbtaches) {
    return -1;
  }
  int res = l->tab[l->debut];
  if (rand()%100==1) {
    usleep(10);
  }
  l->debut++;

  return res;
}

void usage(const char* nom) {
  printf("Usage : %s dim0 dim1 Xmin Ymin Xmax Ymax Prof \n", nom);
}

descr lecture_args(int argc, char *argv[]) {
  descr   d;

  /* Valeurs par defaut de la fractale */
  // 2000 2000 -0.736 -0.184 -0.735 -0.183 500
  d.xmin = -0.736;
  d.ymin = -0.184;
  d.xmax = -0.735;
  d.ymax = -0.183;
  d.dimX = 2000;
  d.dimY = 2000;
  d.prof = 500;
  
  /* Recuperation des parametres */
  if( argc > 1) {
    d.dimX = atoi(argv[1]);
    if (d.dimX == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if ( argc > 2) {
    d.dimY = atoi(argv[2]);
    if (d.dimY == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 3) d.xmin = atof(argv[3]);
  if( argc > 4) d.ymin = atof(argv[4]);
  if( argc > 5) d.xmax = atof(argv[5]);
  if( argc > 6) d.ymax = atof(argv[6]);
  if( argc > 7)  {
    d.prof = atoi(argv[7]);
    if (d.prof == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 8) {
      usage(argv[0]);
      exit(1);
  }

  return d;
}
/// calcul de la couleur d'une ligne
void calcul_ligne(image_t image, descr d, int numligne) {
  int i;
  int depart = numligne*d.dimY;
  double x = numligne*(d.xmax-d.xmin)/d.dimX+d.xmin;
  double y = d.ymin;
  double pas = (d.ymax-d.ymin)/d.dimY;

  if ((numligne<0)||(numligne>=d.dimX)) {
    fprintf(stderr, "Attention, vous tentez de calculer la ligne %d d'une image qui a %d lignes\n",
	    numligne, d.dimX);
    return;
  }
  
  pixel_t *p = image.data+depart;
  if (*p != 0) {
    fprintf(stderr, "Attention, la ligne %d de l'image image semble avoir déjà été calculée (je continue, mais il y a peut-être une erreur)\n",
	    numligne);    
  }
  for (i=0; i<d.dimY; i++) {
    p[i] = xy2color(x,y, d.prof);
    y += pas;
  }
}


// ********************************************************************************
// 
//   modifier à partir de là
//   Partie principale: pour chaque ligne de l'image appliquer la fonction précédante
// ********************************************************************************

int main(int argc, char *argv[]) {
  descr   d;
  clock_t debut, fin;		    /* chronometrage */


  
  /* debut du chronometrage */
  debut = clock();

  /* lecture des argument */
  d = lecture_args(argc, argv);
  
  /* affichage parametres pour control */
  printf( "Domaine: %lf, %lf %lf %lf\n", 
	  d.xmin, d.ymin, d.xmax, d.ymax); 
  printf( "Prof: %d\n",  d.prof);
  printf( "Dim image: %d %d\n", d.dimX, d.dimY);
  
  // création de l'image
  image_t im = image_new(d.dimX, d.dimY);

  // création de la liste des tâches
  lt *l = new_liste_taches(d.dimX);


  // ##############################################
  // Partie à modifier, il faut lancer des threads
  // qui font plus ou moins cette boucle
  // ##############################################
  /* Traitement de l'image ligne par ligne*/
  int nblignes = 0;
  while (1) {
    int t = get_tache(l);
    if (t == -1) {// la liste de taches est terminée et dans ce cas la fonction retourne -1
      break;
    }
    calcul_ligne(im, d, t);
    nblignes++;
  }

  // ################################################
  // a priori ne rien changer ensuite
  // ################################################
  // c'est une vérification pour vous aider a voir les probèmes, cela ne servira qu'en cas de multithreading (et si vous parvenez
  // bien à calculer le nombre de lignes fait par chaque threads.
  if (nblignes == d.dimX) {
    fprintf(stdout, "on a bien calculé %d lignes\n", nblignes);
  } else {
    fprintf(stdout, "Attention, %d lignes ont été calculées à la place de %d (ou  le calcul du nombre de ligne est faux)\n", nblignes, d.dimX);
  }
  
  /* Sauvegarde de la grille dans le fichier resultat "result.pgm" */
  image_to_file(im, "result.pgm");
  
  /* fin du chronometrage */
  fin = clock();
  printf("Temps total de calcul : %e seconde \n",
		 (double) (fin - debut)/ CLOCKS_PER_SEC);

  delete_liste_taches(l);
  
  return 0;
}
